#!/usr/bin/env ruby
# -*- coding: utf-8 -*-
# alone : web application framework for small embedded systems.
#               Copyright (c) 2009-2010 Inas Co Ltd. All Rights Reserved.
#
# This file is destributed under BSD License. Please read the COPYRIGHT file.
#
# データ永続化機構
#


##
# データ永続化 ベースクラス
#
class AlPersist
  ##
  # Marshalベースファイル永続化 ファクトリ
  #
  #@param [String] fname  ファイルネーム
  #@return [AlPersist]    永続化オブジェクト
  #
  def self.file( fname )
    require 'al_persist_file'
    return AlPersistFile.new( fname )
  end


  ##
  # SQLite3ベース永続化 ファクトリ
  #
  #@param [String] fname  ファイルネーム
  #@return [AlPersist]    永続化オブジェクト
  #
  def self.sqlite( fname )
    require 'al_persist_sqlite'
    return AlPersistSqlite.connect( fname )
  end


  ##
  # PostgreSQLベース永続化 ファクトリ
  #
  #@param [String] conn_str   接続文字列
  #@return [AlPersist]    永続化オブジェクト
  #
  def self.postgres( conn_str )
    require 'al_persist_postgres'
    return AlPersistPostgres.connect( conn_str )
  end


  ##
  # RDB wrapperオブジェクトを得る
  #
  #@param [String] conn_str     接続文字列
  #@return [AlRdbw]             RDB wrapper オブジェクト
  #@note
  # 既に接続は完了していて、その接続オブジェクトを得る場合に使用する。
  # DBの種類などが明らかにならないので、いきなり使うことはできない。
  #
  def self.connect( conn_str = nil )
    return AlRdbw.connect( conn_str )
  end


  #@return [AlRdbw] 使用する RDB wrapper オブジェクト
  attr_reader :persist_base

  #@return [Hash] 管理するアトリビュート
  attr_accessor :values

  #@return [Array<Symbol>] プライマリキーの配列
  attr_reader :pkeys

  #@return [Hash] 検索条件のキャッシュ
  attr_reader :search_condition


  ##
  # constructor
  #
  #@param [AlRdbw] base         使用する RDB wrapper オブジェクト
  #@param [Array<String,Symbol>,String,Symbol]  keys   プライマリキー
  #
  def initialize( base, keys = nil )
    @persist_base = base
    @values = {}
    pkey( keys )
    @search_condition = {}
  end


  ##
  # primary key setter
  #
  #@param [Symbol,String]       keys    プライマリキー
  #@return [self]       selfオブジェクト
  #
  def pkey( *keys )
    @pkeys = []

    keys.flatten.each do |k|
      case k
      when String
        @pkeys << k.to_sym
      when Symbol
        @pkeys << k
      when NilClass
        # nothing
      else
        raise 'Needs key by String or Symbol'
      end
    end

    return self
  end


  ##
  # attribute getter
  #
  #@param [Symbol] k    キー
  #@return [String]     値
  #
  def []( k )
    return @values[k.to_sym]
  end


  ##
  # attribute setter
  #
  #@param [Symbol] k    キー
  #@param [String] v    値
  #
  def []=( k, v )
    @values[k.to_sym] = v
  end


  ##
  # 検索時、次ページのオフセット値を得る
  #
  #@return [Integer,Nil]        オフセット値。次ページがなければnil。
  #
  def get_next_offset()
    total_rows = @search_condition[:total_rows].to_i
    num_rows = @search_condition[:num_rows].to_i
    limit = @search_condition[:limit].to_i
    offset = @search_condition[:offset].to_i

    return (offset + num_rows) < total_rows ? (offset + limit) : nil
  end


  ##
  # 検索時、前ページのオフセット値を得る
  #
  #@return [Integer,Nil]        オフセット値。前ページがなければnil。
  #
  def get_previous_offset()
    limit = @search_condition[:limit].to_i
    offset = @search_condition[:offset].to_i

    return offset > 0 ? (offset - limit) : nil
  end

end
