#!/usr/bin/env ruby
# -*- coding: utf-8 -*-
# alone : web application framework for small embedded systems.
#               Copyright (c) 2009-2010 Inas Co Ltd. All Rights Reserved.
#
# This file is destributed under BSD License. Please read the COPYRIGHT file.
#

require 'al_persist'


##
# データ永続化 RDB共通クラス
#
class AlPersistRDB < AlPersist

  #@return [String]  テーブル名
  attr_reader :table


  ##
  # constructor
  #
  #@param [AlRdbw] base         使用する RDB wrapper オブジェクト
  #@param [String] tname        テーブル名
  #@param [Array<String,Symbol>,String,Symbol]  keys   プライマリキー
  #
  def initialize( base, tname, keys = nil )
    super( base, keys )
    @table = tname
  end


  ##
  # キーを指定してデータを読み込み、内部(@values) に保持する。
  #
  #@param  [Hash]       values  プライマリキーを含むHash
  #@return [Boolean]    成功／失敗
  #@note
  # valuesには、プライマリキー以外の値が含まれていてもよく、単に無視される。
  #
  def read( values = nil )
    @values = values  if values
    @search_condition = {}
    return false  if @values.empty?

    #
    # exclude keys
    #
    val = {}
    if ! @pkeys.empty?
      @pkeys.each do |k|
        return false  if ! @values.key?( k )
        val[k] = @values[k]
      end
    else
      @values.each do |k,v|
        val[k.to_sym] = v
      end
    end

    #
    # run sql
    #
    rows = @persist_base.select( "select * from #{@table} _WHERE_;", val )
    return false  if rows.empty?

    @values = rows[0]
    return true
  end


  ##
  # データをRDBへ、新規保存する。
  #
  #@param  [Hash]       values プライマリキーを含む保存する値のHash
  #@return [Boolean]    成功／失敗
  #@note
  # 引数の指定があれば、その値を一旦内部値(@values)にしたうえで、
  # 内部値を新規保存する。
  #
  def create( values = nil )
    @values = values  if values

    res = @persist_base.insert( @table, @values )
    return res[:cmdtuples] == 1 ? true : false
  end


  ##
  # キーの合致するRDB内のデータを、更新する。
  #
  #@param  [Hash]       values プライマリキーを含む更新する値のHash
  #@return [Boolean]    成功／失敗
  #@note
  # 引数の指定があれば、その値を一旦内部値(@values)にしたうえで更新する。
  #
  def update( values = nil )
    @values = values  if values
    return false  if @pkeys.empty?

    value_hash = @values.dup
    where_hash = {}

    # exclude key
    @pkeys.each do |k|
      raise "No key included in value. #{k}"  if ! value_hash[k]
      where_hash[k] = value_hash[k]
      value_hash.delete( k )
    end

    res = @persist_base.update( @table, value_hash, where_hash )
    return res[:cmdtuples] == 1 ? true : false
  end


  ##
  # キーの合致するRDB内のデータを、削除する。
  #
  #@param  [Hash]       values プライマリキーを含むHash
  #@return [Boolean]    成功／失敗
  #@note
  # valuesには、プライマリキー以外の値が含まれていてもよく、単に無視される。
  # 引数の指定があれば、その値を一旦内部値(@values)にしたうえで削除する。
  #
  def delete( values = nil )
    @values = values  if values

    if @pkeys.empty?
      res = @persist_base.delete( @table, @values )
    else
      where_hash = {}

      # exclude keys
      @pkeys.each do |k|
        raise "No key included in value. #{k}"  if ! @values[k]
        where_hash[k] = @values[k]
      end

      res = @persist_base.delete( @table, where_hash )
    end

    return res[:cmdtuples] == 1 ? true : false
  end


  ##
  # キーの合致するデータがあれば更新し、なければ新規登録する。
  #
  #@param  [Hash]       values プライマリキーを含む保存する値のHash
  #@return [Boolean]    成功／失敗
  #@note
  # 引数の指定があれば、その値を一旦内部値(@values)にしたうえで登録する。
  #
  def entry( values = nil )
    delete( values )
    create()
  end


  ##
  # 登録データをすべて読み込む。
  #
  #@return [Array<AlPersist>]     AlPersistオブジェクトの配列
  #@note
  # データ1件を1つのAlPersistオブジェクトとして配列で返す。
  # 一件もデータがない場合は、空の配列を返す。
  # read()と違って自分は何も変わらず、自分の複製を生産する。
  #
  def all()
    rows = @persist_base.select( "select * from #{@table};" )

    ret = []
    rows.each do |row|
      a = self.dup
      a.values = row
      ret << a
    end

    @search_condition = { :total_rows=>ret.size, :num_rows=>ret.size }
    return ret
  end


  ##
  # 登録データを選択的に読み込む。
  #
  #@param  [Hash]  param                検索パラメータ
  #@option param [Hash]  :where         検索条件のハッシュ
  #@option param [Hash,Array,String] :order_by  並べ替え順
  #@option param [Integer] :limit       最大取得数
  #@option param [Integer] :offset      オフセット
  #@option param [Integer] :total_rows  全件数のキャッシュ
  #@return [Array<AlPersist>]           Persistオブジェクトの配列
  #@see AlRdbwSqlite#select
  #@note
  # データ1件を1つのAlPersistオブジェクトとして配列で返す。
  # 一件もデータがない場合は、空の配列を返す。
  # read()と違って自分は何も変わらず、自分の複製を生産する。
  # :total_rowsが指定なき時は、全件数を取得してから結果を取得する。
  # これが速度的に問題になるときは、ダミー値を入れておく。
  #
  def search( param = {} )
    #
    # Where option
    #
    if param[:where]
      @search_condition[:where] = param[:where]
    elsif param.has_key?(:where)
      @search_condition.delete(:where)
    end

    #
    # SQL order by
    #
    sql = ""
    if param[:order_by]
      @search_condition[:order_by] = param[:order_by]
    elsif param.has_key?(:order_by)
      @search_condition.delete(:order_by)
    end
    if @search_condition[:order_by]
      case @search_condition[:order_by]
      when Array
        flag_1st = true
        @search_condition[:order_by].each do |k|
          sql << (flag_1st ? " order by #{k}" : ",#{k}")
          flag_1st = false
        end

      when Hash
        flag_1st = true
        @search_condition[:order_by].each do |k,v|
          next if /^asc|desc$/i !~ v
          sql << (flag_1st ? " order by #{k} #{v}" : ",#{k} #{v}")
          flag_1st = false
        end

      when String
        sql << " order by #{@search_condition[:order_by]}"
      end
    end

    #
    # SQL limit, offset
    #
    if param[:limit]
      limit = param[:limit].to_i
      limit = 1  if limit < 1
      @search_condition[:limit] = limit
    elsif param.has_key?(:limit)
      @search_condition.delete(:limit)
    end
    if param[:offset]
      offset = param[:offset].to_i
      offset = 0  if offset < 0
      @search_condition[:offset] = offset
    elsif param.has_key?(:offset)
      @search_condition.delete(:offset)
    end
    if @search_condition[:limit]
      sql << " limit #{@search_condition[:limit]}"
      if @search_condition[:offset]
        sql << " offset #{@search_condition[:offset]}"
      end
    end

    #
    # 実データ取得
    #
    if @search_condition[:where]
      sql = "select * from #{@table} _WHERE_ #{sql};"
      rows = @persist_base.select( sql, @search_condition[:where] )
    else
      sql = "select * from #{@table} #{sql};"
      rows = @persist_base.select( sql )
    end

    @search_condition[:num_rows] = rows.count
    ret = []
    rows.each do |row|
      a = self.dup
      a.values = row
      ret << a
    end

    #
    # 件数取得
    #
    if param[:total_rows]
      @search_condition[:total_rows] = param[:total_rows]
    elsif param.has_key?(:total_rows)
      @search_condition.delete(:total_rows)
      return ret
    end
    if ! @search_condition[:limit]
      return ret
    end

    # num_rowsに矛盾がないか確認
    if @search_condition[:total_rows]
      num_rows = @search_condition[:total_rows].to_i - @search_condition[:offset].to_i
      if num_rows > param[:limit]
        num_rows = param[:limit]
      end

      if @search_condition[:num_rows] == num_rows
        return ret
      end
    end

    # 件数の取得
    if @search_condition[:where]
      sql = "select count(*) as numrows from #{@table} _WHERE_;"
      rows = @persist_base.select( sql, @search_condition[:where] )
    else
      sql = "select count(*) as numrows from #{@table};"
      rows = @persist_base.select( sql )
    end
    @search_condition[:total_rows] = rows[0][:numrows].to_i

    return ret
  end

end
